/************************************************
 Obsluga wyswietlacza LCD w trybie 4 bity
 RC-51 Raisonance
 Przykladowy program w jezyku C dla EP 
-------------------------------------------------
 jacek.bogusz@ep.com.pl
*************************************************/
// wybr modelu pamieci	
#pragma SMALL
// dolaczenie definicji rejestrw '51
#include <reg51.h>

// definicje znakw specjalnych dla wyswietlacza LCD
char code CGRom[65] = {0xAA,0x55,0xAA,0x55,0xAA,0x55,0xAA,0x55,		// "kratka"	0x00
			0xC0,0xC0,0xFF,0xF1,0xF1,0xF1,0xFF,0xC0,		// pusty kwadrat	0x01
			0xC0,0xC0,0xFF,0xFF,0xFF,0xFF,0xFF,0xC0,		// kwadrat zacz.	0x02
			0xE0,0xFF,0xFF,0xFF,0xE0,0xE0,0xE0,0xE0,		// linia		0x03
			0xFF,0xFF,0xF9,0xF3,0xE7,0xF3,0xF9,0xFF,		// znak w lewo	0x04
			0xFF,0xFF,0xF3,0xF9,0xFC,0xF9,0xF3,0xFF,		// znak w prawo	0x05
			0xFF,0xFF,0xFB,0xF1,0xE4,0xEE,0xFF,0xFF,		// znak w gre	0x06
			0xFF,0xFF,0xFF,0xEE,0xE4,0xF1,0xFB,0xFF,		// znak w dl	0x07
			0x00};

// port,do ktrego podlaczono wyswietlacz LCD
#define	PORT		P2
// 4 najstarsze bity, to bity danych, 4 mlodsze moga byc w dowolny sposb skonfigurowane jako sterujace
// po zmianie definicji moga to byc rwniez 2 rzne porty,jednak w programie przykladowym zakladam 
// wykorzystanie tylko jednego z portw mikrokontrolera

// bity sterujace LCD
sbit LcdEnable = 		PORT^0;
sbit LcdRead = 		PORT^3;
sbit LcdReg = 		PORT^1;


// opznienie okolo 1 milisekundy dla kwarcu 7,3728MHz
void Delay (unsigned int k) 	
{
	unsigned int i,j;
	for (j = 0; j < k; j++)
		for (i = 0; i <= 296; i++);
}


// zapis bajtu do lcd, osobna procedura dla zmniejszenia objetosci kodu wynikowego
void WriteByteToLcd(char X)
{
	LcdEnable = 1;
	PORT |= 0xF0;					// ustawienie grnej polwki portu PORT na "1"
	PORT &= (X | 0x0F);				// "bezkolizyjny" zapis 1-szej polwki bajtu (przez funkcje logiczna)
	LcdEnable = 0;					// zapis do wyswietlacza (opadajace zbocze sygnalu E)
	
	LcdEnable = 1;					// zapis 2-giej polwki bajtu
	X <<= 4;					// przesuniecie 4x w lewo
	PORT |= 0xF0;					// ustawienie grnej polwki portu PORT na "1"
	PORT &= (X | 0x0F);				// zapis 2-giej polwki bajtu i maskowanie 4 mlodszych bitw
	LcdEnable = 0;					// opadajace zbocze E - zapis do LCD

	Delay(1);
}

// zapis bajtu do rejestru kontrolnego LCD
void WriteToLcdCtrlRegister(char X)
{
	LcdReg = LcdRead = 0;				// ustawienie sygnalw sterujacych
	WriteByteToLcd(X);
}


// zapis bajtu do pamieci obrazu
void LcdWrite(char X)
{
	LcdReg = 1;					// ustawienie sygnalw sterujacych
	LcdRead = 0;
	LcdEnable = 1;
	WriteByteToLcd(X);
}


// czyszczenie ekranu LCD
void LcdClrScr(void)
{
	WriteToLcdCtrlRegister(0x01);
}


// inicjalizacja wyswietlacza LCD w trybie 4 bity
void LcdInitialize(void)
{
	char i;
	
	Delay(15);
	P0 = 0x0F;					// wyzerowanie linii LcdReg,LcdRead,LcdEnable
	for (i = 0; i<3; i++)
	{
		LcdEnable = 1;				// impuls na E
		PORT &= 0x3F;				// ustawienie wartosci inicjujacej
		LcdEnable = 0;
		Delay(5);
	}
	LcdEnable = 1;					// wpisanie wartosci 2 do rejestru kontrolnego
	PORT &= 0x2F;					// tylko "grne" 4 bity
	LcdEnable = 0;
	Delay(1);
	WriteToLcdCtrlRegister(0x28);			// interfejs 4 bity,znaki 5x7
	WriteToLcdCtrlRegister(0x08);			// wylaczenie LCD
	WriteToLcdCtrlRegister(0x01);			// kasowanie ekranu, powrt do pozycji home
	WriteToLcdCtrlRegister(0x06);			// przesuwanie kursora z inkrementacja
	WriteToLcdCtrlRegister(0x0C);			// zalaczenie wyswietlacza
}

	
// ustawia kursor na wsplrzednych x,y
void GotoXY(char x, char y)
{
	switch (y)					// obliczenie o ile nalezy zwiekszyc x w zaleznosci od wartosci y 
	{
		case 0:
			x += 0x80;			// x = x + 0x80
			break;
		case 1:
			x += 0xC0;			// x = x + 0xC0
			break;
		case 2:
			x += 0x94;			// x = x + 0x94
			break;
		case 3:
			x += 0xD4;			// x = x + 0xD4
			break;
	}
	WriteToLcdCtrlRegister(x);
}


// wywietla tekst na wsprzdnych x, y
void WriteTextXY(char x, char y, char *S)
{
	while (*S)					// petla dziala dotad, az napotkany zostanie znak konca lancuch (/0)
	{
		GotoXY(x, y);				// wyliczenie adresu dla znaku
		LcdWrite(*S);				// wyswietlenie pojedynczego znaku
		x++; 					// nastepna pozycja x na ekranie
		S++;					// nastepna pozycja wskaznika, tzn. nastepny znak napisu
		if (x > 19)				// jesli x>19 to nastepna linia
		{
			x = 0; 
			y++;
		}
	}
}


// wywietla tekst rozpoczynajc od aktualnego pooenia kursora
void WriteText(char *S)
{
	while (*S)					// petla dziala dotad,az napotkany zostanie znak konca lancucha (/0)
	{
		LcdWrite(*S);				// wyswietlenie pojedynczego znaku
		S++;
	}
}


// definiowanie znakw z tablicy CGRom
void DefineSpecialCharacters(char *ptr)
{
	WriteToLcdCtrlRegister(0x40);			// ustawienie trybu definicji
	while (*ptr != 0)					// petla wykonywana do napotkania znaku konca tablicy
	{
		LcdWrite(*ptr);				// zapis znaku do lcd cgram
		ptr++;					// nastepna pozycja tablicy (wskaznika)
	}
	WriteToLcdCtrlRegister(0x80);			// przelaczenie do trybu wyswietlania
}

// program glwny
void main(void)
{
	char ix = 1, iy = 1, x, y, i = 0;
	
	LcdInitialize();
	DefineSpecialCharacters(&CGRom);
	while (1)
	{
		WriteTextXY(x, y, " ");
		if (ix == 1) x++; else x--;
		if (iy == 1) y++; else y--;
		if (x == 19) ix = 0;
		if (x == 0) ix = 1;
		if (y == 3) iy = 0;
		if (y == 0) iy = 1;
		WriteTextXY(x, y, 0x01);
		Delay(50);
	}
}

